//---------------------------------------------------------------------------
// FT1000 driver for Flarion Flash OFDM NIC Device
//
// Copyright (C) 2006 Flarion Technologies, All rights reserved.
//
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option) any 
// later version. This program is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
// more details. You should have received a copy of the GNU General Public 
// License along with this program; if not, write to the 
// Free Software Foundation, Inc., 59 Temple Place - 
// Suite 330, Boston, MA 02111-1307, USA. 
//---------------------------------------------------------------------------
//
// File:         ft1000_chdev.c
//
// Description:  Custom character device dispatch routines.
//
// History:
// 8/29/02    Whc                Ported to Linux.
// 6/05/06    Whc                Porting to Linux 2.6.9
//
//---------------------------------------------------------------------------
#include <linux/config.h>

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/poll.h>
#include <linux/netdevice.h>
#include <linux/delay.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/fs.h>
#include <linux/kmod.h>
#include <linux/ioctl.h>

#include "ft1000.h"
#include "ft1000_ioctl.h"

extern inline u16 ft1000_asic_read (struct net_device *dev, u16 offset);
extern inline void ft1000_asic_write (struct net_device *dev, u16 offset, u16 value);
extern void ft1000_send_cmd (struct net_device *dev, u16 *ptempbuffer, int size, u16 qtype);

static int ft1000_ChOpen (struct inode *Inode, struct file *File); 
static int ft1000_ChIoctl (struct inode *Inode, struct file *File, unsigned int Command,
                           unsigned long Argument); 
static int ft1000_ChRelease (struct inode *Inode, struct file *File);

static int ft1000_flarion_cnt = 0;

// Global pointer to device object
static struct net_device *pdevobj;

// Global variable to indicate that all provisioning data is sent to DSP
BOOLEAN fProvComplete;

//
// Table of entry-point routines for char device
//
static struct file_operations ft1000fops =
{
    ioctl:    ft1000_ChIoctl,
    open:     ft1000_ChOpen,        
    release:  ft1000_ChRelease
};

//---------------------------------------------------------------------------
// Function:    exec_mknod
//
// Parameters:  
//
// Returns:
//
// Description:
//
// Notes:
//
//---------------------------------------------------------------------------
static int exec_mknod (void *pdata) 
{
    PFT1000_INFO info;
    char *argv[]={"mknod","-m 666","/dev/FT1000","c","000","000","NULL"};
    int retcode;
    char mjnum[4];

    info = pdata;
    DEBUG(1,"ft1000_chdev:exec_mknod is called with major number = %d\n", info->DeviceMajor);
    argv[0] = "mknod";
    argv[1] = "-m 666";    
    argv[2] = "/dev/FT1000";
    argv[3] = "c";
    sprintf(mjnum, "%d", info->DeviceMajor);
    argv[4] = mjnum;
    argv[5] = "0";
    argv[6] = NULL;

    retcode = call_usermodehelper ("/bin/mknod", argv, NULL, 1);
    if (retcode) {
        DEBUG(1,"ft1000_chdev:exec_mknod failed to make the node: retcode = %d\n", retcode);
    }

    return retcode;

}

//---------------------------------------------------------------------------
// Function:    rm_mknod
//
// Description: This module removes the FT1000 device file
//
//---------------------------------------------------------------------------
static int rm_mknod (void *pdata) 
{
    char *argv[4];
    int retcode;

    DEBUG(1,"ft1000_chdev:rm_mknod is called\n");
    argv[0] = "rm";
    argv[1] = "-f";    
    argv[2] = "/dev/FT1000";
    argv[3] = NULL;

    retcode = call_usermodehelper ("/bin/rm", argv, NULL, 1);
    if (retcode) {
        DEBUG(1,"ft1000_chdev:rm_mknod failed to remove the node: retcode = %d\n", retcode);
    }

    return retcode;

}

//---------------------------------------------------------------------------
// Function:    ft1000_CreateDevice
//
// Parameters:  dev - pointer to adapter object
//
// Returns:     0 if successful
//
// Description: Creates a private char device.
//
// Notes:       Only called by init_module().
//
//---------------------------------------------------------------------------
int ft1000_CreateDevice(struct net_device *dev)
{
    PFT1000_INFO info = (PFT1000_INFO)dev->priv;
    int result;
    int i;
    pid_t pid;

    // Delete any existing FT1000 node
    pid = kernel_thread (rm_mknod,(void *)info, 0);
    
    DEBUG(1,"ft1000_CreateDevice: number of instance = %d\n", ft1000_flarion_cnt);
    DEBUG(1,"DeviceCreated = %x\n", info->DeviceCreated);
    pdevobj = dev;
    if (info->DeviceCreated)
    {
	DEBUG(1, "ft1000_CreateDevice: \"%s\" already registered\n", info->DeviceName);
	return -EIO;
    }

    // make a new device name
    sprintf(info->DeviceName, "%s", "FT1000");

    // register the device
    DEBUG(1, "ft1000_CreateDevice: \"%s\" device registration\n", info->DeviceName);
    info->DeviceMajor = 0;
	
    result = register_chrdev(info->DeviceMajor, info->DeviceName, &ft1000fops);
    if (result < 0)
    {
	DEBUG(1,"ft1000_CreateDevice: unable to get major %d\n", info->DeviceMajor);
	return result;
    }

    DEBUG(1, "ft1000_CreateDevice: registered char device \"%s\"\n", info->DeviceName);

    // save a dynamic device major number
    if (info->DeviceMajor == 0)
    {
	info->DeviceMajor = result;
	DEBUG(1,"ft1000_PcdCreateDevice: device major = %d\n", info->DeviceMajor);
    }

    // Create a thread to call user mode app to mknod
    pid = kernel_thread (exec_mknod, (void *)info, 0);

    // initialize application information 
    info->appcnt = 0;
    for (i=0; i<MAX_NUM_APP; i++) {
        info->app_info[i].fileobject = 0;
        info->app_info[i].app_id = i+1;
    }

    info->DeviceCreated = TRUE;
    ft1000_flarion_cnt++;

    return result;
}

//---------------------------------------------------------------------------
// Function:    ft1000_DestroyDevice
//
// Parameters:  dev - pointer to adapter object
//
// Description: Destroys a private char device.
//
// Notes:       Only called by cleanup_module().
//
//---------------------------------------------------------------------------
void ft1000_DestroyDevice(struct net_device *dev)
{
    PFT1000_INFO info = (PFT1000_INFO)dev->priv;
    int result;
    pid_t pid;

    DEBUG(1,"ft1000_chdev:ft1000_DestroyDevice called\n");

    if (info->DeviceCreated)
    {
        ft1000_flarion_cnt--;		
	result = unregister_chrdev(info->DeviceMajor, info->DeviceName);
	DEBUG(1,"ft1000_DestroyDevice: unregistered device %s result = %d\n", 
        info->DeviceName, result);

        pid = kernel_thread (rm_mknod, (void *)info, 0);

        info->DeviceCreated = FALSE;
    }
}

//---------------------------------------------------------------------------
// Function:    ft1000_ChOpen
//
// Parameters:  
//
// Description: 
//
// Notes:
//
//---------------------------------------------------------------------------
static int ft1000_ChOpen (struct inode *Inode, struct file *File) 
{
    PFT1000_INFO info;
    int i;
    
    DEBUG(1,"ft1000_ChOpen called\n");
    info = (PFT1000_INFO)pdevobj->priv;

    DEBUG(1,"f_owner = 0x%8x number of application = %d\n", (u32)(&File->f_owner), info->appcnt );

    // Check if maximum number of application exceeded
    if (info->appcnt > MAX_NUM_APP) {
        DEBUG(1,"Maximum number of application exceeded\n");
        return -EACCES;
    }

    // Search for available application info block
    for (i=0; i<MAX_NUM_APP; i++) {
        if ( (info->app_info[i].fileobject == 0) ) {
            break;
        }
    }

    // Fail due to lack of application info block
    if (i == MAX_NUM_APP) {
        DEBUG(1,"Could not find an application info block\n");
        return -EACCES;
    }

    info->appcnt++;
    info->app_info[i].fileobject = (u32)(&File->f_owner);
    info->app_info[i].nTxMsg = 0;
    info->app_info[i].nRxMsg = 0;
    info->app_info[i].nTxMsgReject = 0;
    info->app_info[i].nRxMsgMiss = 0;

    File->private_data = pdevobj;

    return 0;
}

//---------------------------------------------------------------------------
// Function:    ft1000_ChIoctl
//
// Parameters:  
//
// Description: 
//
// Notes:
//
//---------------------------------------------------------------------------
static int ft1000_ChIoctl (struct inode *Inode, struct file *File, unsigned int Command,
                           unsigned long Argument) 
{
    struct net_device *dev;
    PFT1000_INFO info;
    int result=0;
    int cmd;
    struct timeval tv;
    IOCTL_GET_VER get_ver_data;
    IOCTL_GET_DSP_STAT get_stat_data;
    u8 ConnectionMsg[] = {0x00,0x44,0x10,0x20,0x80,0x00,0x00,0x00,0x00,0x00,0x03,0x00,0x00,0x00,0x93,0x64,
                          0x00,0x00,0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x05,0x00,0x00,0x00,0x0a,
                          0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
                          0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x12,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
                          0x00,0x00,0x02,0x37,0x00,0x00,0x00,0x08,0x00,0x00,0x00,0x01,0x00,0x01,0x7f,0x00,
                          0x00,0x01,0x00,0x00};
    
    DEBUG(1,"ft1000_ChIoctl called\n");

    if (ft1000_flarion_cnt == 0) {
        DEBUG(1,"FT1000:ft1000_ChIoctl called when ft1000_flarion_cnt is zero\n");
        return (-EBADF);
    }

    DEBUG(1,"FT1000:ft1000_ChIoctl:Command = 0x%x Argument = 0x%8x\n", Command, (u32)Argument);

    dev = File->private_data;
    info = (PFT1000_INFO)dev->priv;
    cmd = _IOC_NR(Command);
    DEBUG(1,"FT1000:ft1000_ChIoctl:cmd = 0x%x\n", cmd);
   
    // process the command
    switch (cmd) {
    case IOCTL_GET_VER_CMD:
        DEBUG(1,"FT1000:ft1000_ChIoctl: IOCTL_FT1000_GET_VER called\n");

        get_ver_data.drv_ver = FT1000_DRV_VER;

        if (copy_to_user((PIOCTL_GET_VER)Argument, &get_ver_data, sizeof(get_ver_data)) ) {
            DEBUG(1,"FT1000:ft1000_ChIoctl: copy fault occurred\n");
            result = -EFAULT;
            break;
        }

        DEBUG(1,"FT1000:ft1000_ChIoctl:driver version = 0x%x\n",(unsigned int)get_ver_data.drv_ver);

        break;
    case IOCTL_CONNECT:
        // Connect Message
        DEBUG(0,"FT1000:ft1000_ChIoctl: IOCTL_FT1000_CONNECT\n");
        ConnectionMsg[79] = 0xfc;
        ft1000_send_cmd (dev, (u16 *)ConnectionMsg, sizeof(ConnectionMsg), 0);
        break;
    case IOCTL_DISCONNECT:
        // Disconnect Message
        DEBUG(0,"FT1000:ft1000_ChIoctl: IOCTL_FT1000_DISCONNECT\n");
        ConnectionMsg[79] = 0xfd;
        ft1000_send_cmd (dev, (u16 *)ConnectionMsg, sizeof(ConnectionMsg), 0);
        break;
    case IOCTL_GET_DSP_STAT_CMD:
        DEBUG(1,"FT1000:ft1000_ChIoctl: IOCTL_FT1000_GET_DSP_STAT called\n");

        memcpy(get_stat_data.DspVer, info->DspVer, DSPVERSZ);
        memcpy(get_stat_data.HwSerNum, info->HwSerNum, HWSERNUMSZ);
        memcpy(get_stat_data.Sku, info->Sku, SKUSZ);
        memcpy(get_stat_data.eui64, info->eui64, EUISZ);

        if (info->AsicID == ELECTRABUZZ_ID) {
            get_stat_data.LedStat = ft1000_read_dpram (dev, FT1000_DSP_LED);
            if (info->ProgConStat != 0xFF) {
                get_stat_data.ConStat = ft1000_read_dpram (dev, FT1000_DSP_CON_STATE);
            }
            else {
                get_stat_data.ConStat = 0x0f;
            }
        }
        else {
            if (info->ProgConStat != 0xFF) {
                get_stat_data.LedStat = ntohs(ft1000_read_dpram_mag_16 (pdevobj, FT1000_MAG_DSP_LED, FT1000_MAG_DSP_LED_INDX));
                DEBUG(1,"FT1000:ft1000_ChIoctl: LedStat = 0x%x\n", get_stat_data.LedStat);
                get_stat_data.ConStat = ntohs(ft1000_read_dpram_mag_16 (pdevobj, FT1000_MAG_DSP_CON_STATE, FT1000_MAG_DSP_CON_STATE_INDX));
                DEBUG(1,"FT1000:ft1000_ChIoctl: ConStat = 0x%x\n", get_stat_data.ConStat);
            }
            else {
                get_stat_data.ConStat = 0x0f;
            }
        }
            
        get_stat_data.nTxPkts = info->stats.tx_packets;
        get_stat_data.nRxPkts = info->stats.rx_packets;
        get_stat_data.nTxBytes = info->stats.tx_bytes;
        get_stat_data.nRxBytes = info->stats.rx_bytes;
        do_gettimeofday ( &tv );
        get_stat_data.ConTm = (u32)(tv.tv_sec - info->ConTm);
        DEBUG(1,"Connection Time = %d\n", (int)get_stat_data.ConTm); 
        if (copy_to_user((PIOCTL_GET_DSP_STAT)Argument, &get_stat_data, sizeof(get_stat_data)) ) {
            DEBUG(1,"FT1000:ft1000_ChIoctl: copy fault occurred\n");
            result = -EFAULT;
            break;
        }

        break;    

    default:
        DEBUG(1,"FT1000:ft1000_ChIoctl:unknown command: 0x%x\n", Command);
        result = -ENOTTY;
        break;
    }

    return result;
}

//---------------------------------------------------------------------------
// Function:    ft1000_ChRelease
//
// Parameters:  
//
// Description: 
//
// Notes:
//
//---------------------------------------------------------------------------
static int ft1000_ChRelease (struct inode *Inode, struct file *File)
{
    PFT1000_INFO info;
    struct net_device *dev;
    int i;
    
    DEBUG(1,"ft1000_ChRelease called\n");

    dev = File->private_data;
    info = (PFT1000_INFO)dev->priv;

    if (ft1000_flarion_cnt == 0) {
        info->appcnt--;
        return (-EBADF);
    }

    // Search for matching file object
    for (i=0; i<MAX_NUM_APP; i++) {
        if ( info->app_info[i].fileobject == (u32)(&File->f_owner) ) {
            DEBUG(1,"FT1000:ft1000_ChIoctl: Message is for AppId = %d\n", info->app_info[i].app_id);
            break;
        }
    }

    if (i==MAX_NUM_APP)
	    return 0;

    
    // initialize application information 
    info->appcnt--;
    DEBUG(1,"ft1000_chdev:%s:appcnt = %d\n", __FUNCTION__, info->appcnt);
    info->app_info[i].fileobject = 0;
    
    return 0;
}

