//---------------------------------------------------------------------------
//
// File:         getstat.c
//
//
// Description:    Monitor Application to retrieve Connect Status, Signal Strength,
//                 and Signal Quality from FT1000 device driver.
//
// History:
// 11/5/02    Whc                Created.
//
// Copyright (C)2002 Flarion Technologies.  All Rights Reserved
//---------------------------------------------------------------------------

#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <linux/fs.h>
#include <sys/ioctl.h>
#include <asm/poll.h>

#include "ft1000_ioctl.h"

int main(int argc, char *argv[])
{
    int fd;
    int result; 
    unsigned short app_id;
    IOCTL_GET_DSP_STAT data;
    IOCTL_GET_VER version;
    union {
        unsigned long whole;
        unsigned char byte[4];
    } value;
    int i;
    int toggle;
    unsigned short temp;
    int hour, min, sec;

    fd = open("/dev/FT1000", O_RDWR);
    
    if (fd < 0) {
        // Let's try FT1001 
       fd = open("/dev/FT1001", O_RDWR);
       if (fd < 0) {
           // Let's try FT1002
           fd = open("/dev/FT1002", O_RDWR);
           if (fd < 0) {
	       fd = open("/dev/FT1003", O_RDWR);
	       if (fd < 0) {
                  printf("Device Not Found - Terminating\n");
                  sleep(3);
                  exit(0);
	       }
           }
       }
    }
    
    while (1) {
        system("clear");
        if (toggle) {
           printf("FT1000 Monitor 1.1 -\n");
           toggle = 0;
        }
        else {
           printf("FT1000 Monitor 1.1 |\n");
           toggle = 1;
        }            
        
        result = ioctl(fd, IOCTL_FT1000_GET_VER, &version);
        if (result == -1) {
            if (errno == EBADF) {
                printf("Device Not Found - Terminating\n");
                close(fd);                
                sleep(3);
                exit(0);
            }
            else if (errno == EACCES) {
                printf("Device Not Ready to be access\n");
                sleep(2);
                continue;
            }
            else {
                printf("Unexpected Errno = %d\n", errno);
                close(fd);                
                sleep(3);
                exit(0);
            }
        }

        value.whole = ntohl(version.drv_ver);

        result = ioctl(fd, IOCTL_FT1000_GET_DSP_STAT, &data);
        if (result == -1) {
            if (errno == EBADF) {
                printf("Device Not Found - Terminating\n");
                close(fd);                
                sleep(3);
                exit(0);
            }
            else if (errno == EACCES) {
                printf("Device Not Ready to be access\n");
                sleep(2);
            }
            else {
                printf("Unexpected Errno = %d\n", errno);
                close(fd);                
                sleep(3);
                exit(0);
            }
        }
        else {
            printf("FT1000 linux device driver version = %d.%d.%d.%d\n", value.byte[0], value.byte[1], value.byte[2], value.byte[3]);
            printf("Firmware Version = ");
            for (i=1; i<4; i++) {
                printf("%d", data.DspVer[i]);
                if (i!=3) {
                    printf(".");
                }
            }
            printf("\n");
            printf("EUI-64 = ");
            for (i=0; i<8; i++) {
                printf(" 0x%x ", data.eui64[i]);
            }
            printf("\n");

            if ( (data.ConStat & 0xf) == 2) {
                printf("Connected...\n");
                temp = (data.LedStat & 0xf);
                if (temp == 0) {
                    printf("   Signal Strength = \n");
                }
                else if (temp == 0x1) {
                    printf("   Signal Strength = *\n");
                }
                else if (temp == 0x3) {
                    printf("   Signal Strength = **\n");
                }
                else if (temp == 0x7) {
                    printf("   Signal Strength = ***\n");
                }
                else if (temp == 0xf) {
                    printf("   Signal Strength = ****\n");
                }
                temp = (data.LedStat >> 8) & 0xf;
                if (temp == 0) {
                    printf("   Signal Quality = \n");
                }
                else if (temp == 0x1) {
                    printf("   Signal Quality = *\n");                    
                }
                else if (temp == 0x3) {
                    printf("   Signal Quality = **\n");                    
                }
                else if (temp == 0x7) {
                    printf("   Signal Quality = ***\n");                    
                }
                else if (temp == 0xf) {
                    printf("   Signal Quality = ****\n");                    
                }
                printf("Number of Tx packets = %d\n", data.nTxPkts);
                printf("Number of Rx packets = %d\n", data.nRxPkts);
                hour = (data.ConTm/3600) % 24;
                min = (data.ConTm/60) % 60;
                sec = data.ConTm % 60;
                printf("Connection time = %d:%d:%d\n", hour, min, sec );
            }
            else if ( (data.ConStat & 0xf) == 0xf) {
                printf("Hardware Failure Detected - PC Card disabled\n");
            }
            else {
                printf("Searching...\n");
            }
            sleep(2);
        }
    }
    
    close(fd);

}

